
from pathlib import Path
import sys
sys.path.insert(0, str(Path(__file__).resolve().parent / '..'))

from geofileops import geofile
from geofileops.util import geofileops_gpd

def get_testdata_dir() -> Path:
    return Path(__file__).resolve().parent / 'data'

def test_convexhull_gpkg(tmpdir):
    # Select some data from input to output file
    input_path = get_testdata_dir() / 'parcels.gpkg'
    output_path = Path(tmpdir) / 'parcels.gpkg'
    basetest_convexhull(input_path, output_path)

def test_convexhull_shp(tmpdir):
    # Select some data from input to output file
    input_path = get_testdata_dir() / 'parcels.shp'
    output_path = Path(tmpdir) / 'parcels.shp'
    basetest_convexhull(input_path, output_path)

def basetest_convexhull(input_path, output_path):
    layerinfo_orig = geofile.getlayerinfo(input_path)
    geofileops_gpd.convexhull(
            input_path=input_path,
            output_path=output_path)

    # Now check if the output file is correctly created
    assert output_path.exists() == True
    layerinfo_select = geofile.getlayerinfo(output_path)
    assert layerinfo_orig.featurecount == layerinfo_select.featurecount
    assert len(layerinfo_orig.columns) == len(layerinfo_select.columns)

    output_gdf = geofile.read_file(output_path)
    assert output_gdf['geometry'][0] is not None

def test_buffer_gpkg(tmpdir):
    # Buffer to test dir
    input_path = get_testdata_dir() / 'parcels.gpkg'
    output_path = Path(tmpdir) / 'parcels.gpkg'
    basetest_buffer(input_path, output_path)

def test_buffer_shp(tmpdir):
    # Buffer to test dir
    input_path = get_testdata_dir() / 'parcels.shp'
    output_path = Path(tmpdir) / 'parcels.shp'
    basetest_buffer(input_path, output_path)

def basetest_buffer(input_path, output_path):
    layerinfo_orig = geofile.getlayerinfo(input_path)
    geofileops_gpd.buffer(
            input_path=input_path,
            output_path=output_path,
            distance=1)

    # Now check if the output file is correctly created
    assert output_path.exists() == True
    layerinfo_select = geofile.getlayerinfo(input_path)
    assert layerinfo_orig.featurecount == layerinfo_select.featurecount
    assert len(layerinfo_orig.columns) == len(layerinfo_select.columns)

    output_gdf = geofile.read_file(output_path)
    assert output_gdf['geometry'][0] is not None

def test_simplify_gpkg(tmpdir):
    # Buffer to test dir
    input_path = get_testdata_dir() / 'parcels.gpkg'
    output_path = Path(tmpdir) / 'parcels.gpkg'
    basetest_simplify(input_path, output_path)

def test_simplify_shp(tmpdir):
    # Buffer to test dir
    input_path = get_testdata_dir() / 'parcels.shp'
    output_path = Path(tmpdir) / 'parcels.shp'
    basetest_simplify(input_path, output_path)

def basetest_simplify(input_path, output_path):
    layerinfo_orig = geofile.getlayerinfo(input_path)
    geofileops_gpd.simplify(
            input_path=input_path,
            output_path=output_path,
            tolerance=5)

    # Now check if the tmp file is correctly created
    assert output_path.exists() == True
    layerinfo_select = geofile.getlayerinfo(input_path)
    assert layerinfo_orig.featurecount == layerinfo_select.featurecount
    assert len(layerinfo_orig.columns) == len(layerinfo_select.columns)

def test_dissolve_nogroupby_gpkg(tmpdir):
    # Buffer to test dir
    input_path = get_testdata_dir() / 'parcels.gpkg'
    output_path = Path(tmpdir) / 'parcels.gpkg'
    basetest_dissolve_nogroupby(input_path, output_path)

def test_dissolve_nogroupby_shp(tmpdir):
    # Buffer to test dir
    input_path = get_testdata_dir() / 'parcels.shp'
    output_path = Path(tmpdir) / 'parcels.shp'
    basetest_dissolve_nogroupby(input_path, output_path)

def basetest_dissolve_nogroupby(input_path, output_path):
    geofileops_gpd.dissolve(
            input_path=input_path,
            output_path=output_path,
            explodecollections=True)

    # Now check if the result file is correctly created
    assert output_path.exists() == True
    layerinfo_orig = geofile.getlayerinfo(input_path)
    layerinfo_output = geofile.getlayerinfo(output_path)
    assert layerinfo_output.featurecount == 21
    assert len(layerinfo_output.columns) >= 0

    # Now check the contents of the result file
    input_gdf = geofile.read_file(input_path)
    output_gdf = geofile.read_file(output_path)
    assert input_gdf.crs == output_gdf.crs
    assert len(output_gdf) == layerinfo_output.featurecount
    assert output_gdf['geometry'][0] is not None

def test_dissolve_groupby_gpkg(tmpdir):
    # Buffer to test dir
    input_path = get_testdata_dir() / 'parcels.gpkg'
    output_path = Path(tmpdir) / 'parcels.gpkg'
    basetest_dissolve_groupby(input_path, output_path)

def test_dissolve_groupby_shp(tmpdir):
    # Buffer to test dir
    input_path = get_testdata_dir() / 'parcels.shp'
    output_path = Path(tmpdir) / 'parcels.shp'
    basetest_dissolve_groupby(input_path, output_path)

def basetest_dissolve_groupby(input_path, output_path):
    layerinfo_orig = geofile.getlayerinfo(input_path)
    geofileops_gpd.dissolve(
            input_path=input_path,
            output_path=output_path,
            groupby_columns=['GEWASGROEP'],
            explodecollections=False)

    # Now check if the tmp file is correctly created
    assert output_path.exists() == True
    layerinfo_output = geofile.getlayerinfo(output_path)
    assert layerinfo_output.featurecount == 3
    assert len(layerinfo_orig.columns) == len(layerinfo_output.columns)

    # Now check the contents of the result file
    input_gdf = geofile.read_file(input_path)
    output_gdf = geofile.read_file(output_path)
    assert input_gdf.crs == output_gdf.crs
    assert len(output_gdf) == layerinfo_output.featurecount
    assert output_gdf['geometry'][0] is not None

if __name__ == '__main__':
    #Prepare tempdir
    import tempfile
    import shutil
    tmpdir = Path(tempfile.gettempdir()) / 'test_geofileops_gpd'
    if tmpdir.exists():
        shutil.rmtree(tmpdir)
    if not tmpdir.exists():
        tmpdir.mkdir()

    # Run
    test_dissolve_nogroupby_shp(tmpdir)
    #test_dissolve_nogroupby_gpkg(tmpdir)



import json
import logging
import logging.config
from pathlib import Path
import sys
sys.path.insert(0, str(Path(__file__).resolve().parent / '..'))

from geofileops import geofile

def main():
    
    # Init logging
    script_dir = Path(__file__).resolve().parent
    with open(script_dir / 'logging.json', 'r') as log_config_file:
        log_config_dict = json.load(log_config_file)
    logging.config.dictConfig(log_config_dict)
    #logger = logging.getLogger()
    
    # Go!
    path=r"X:\Monitoring\OrthoSeg\sealedsurfaces\output_vector\sealedsurfaces_BEFL_2019_ofw_18\sealedsurfaces_BEFL_2019_ofw_18.gpkg"
    geofile.create_spatial_index(path=path, layer=None)

if __name__ == '__main__':
    main()




import os
from pathlib import Path
import shutil
import subprocess
import tempfile
from typing import Optional, Tuple

class CTError(Exception):
    def __init__(self, errors):
        self.errors = errors

def create_tempdir(base_dirname: str) -> Path:
    
    defaulttempdir = Path(tempfile.gettempdir())

    for i in range(1, 999999):
        try:
            tempdir = defaulttempdir / f"{base_dirname}_{i:06d}"
            os.mkdir(tempdir)
            return Path(tempdir)
        except FileExistsError:
            continue

    raise Exception(f"Wasn't able to create a temporary dir with basedir: {defaulttempdir / base_dirname}") 

def get_tempfile_locked(
        base_filename: str,
        suffix: str = None,
        dirname: str = None,
        tempdir: Path = None) -> Tuple[Path, Path]:
    # If no dir specified, use default temp dir
    if tempdir is None:
        tempdir = Path(tempfile.gettempdir())
    if dirname is not None:
        tempdir = tempdir / dirname
        if not tempdir.exists():
            os.makedirs(tempdir)

    # Now look for a unique filename based on the base_filename and put a lock file
    for i in range(1, 999999):
        tempfile_path = tempdir / f"{base_filename}_{i:06d}{suffix}"
        tempfilelock_path = tempdir / f"{base_filename}_{i:06d}{suffix}.lock"
        result = create_file_atomic(tempfilelock_path)
        if result is True:
            if not tempfile_path.exists():
                # OK!
                return (tempfile_path, tempfilelock_path)
            else:
                # Apparently the lock file didn't exist, but the file did... so 
                # delete lock file and try again
                tempfilelock_path.unlink()

    raise Exception(f"Wasn't able to create a temporary file with base_filename: {base_filename}, dir: {dir}") 
 
def copyfile(src, dst):
    """
    standard shutil.copyfile is very slow on windows for large files.

    Args:
        src ([type]): [description]
        dst ([type]): [description]
    
    Raises:
        Exception: [description]
    """
    if os.name == 'nt':
        # On windows, this is a lot faster than all shutil alternatives
        #command = f'copy "{src}" "{dst}"'
        command = f'xcopy /j "{src}" "{dst}*"'
        output = ''
        try:
            output = subprocess.check_output(command, shell=True)
        except Exception as ex:
            raise Exception(f"Error executing {command}, with output {output}") from ex
        
    else:
        buffer_size = 1024*1024*5
        with open(src, 'rb') as fsrc, \
             open(dst, 'wb') as fdest:
            shutil.copyfileobj(fsrc, fdest, buffer_size)
    
def copytree(src, dst, symlinks=False, ignore=[]):
    names = os.listdir(src)

    if not os.path.exists(dst):
        os.makedirs(dst)
    errors = []
    for name in names:
        if name in ignore:
            continue
        srcname = os.path.join(src, name)
        dstname = os.path.join(dst, name)
        try:
            if symlinks and os.path.islink(srcname):
                linkto = os.readlink(srcname)
                os.symlink(linkto, dstname)
            elif os.path.isdir(srcname):
                copytree(srcname, dstname, symlinks, ignore)
            else:
                copyfile(srcname, dstname)
            # XXX What about devices, sockets etc.?
        except (IOError, os.error) as ex:
            errors.append((srcname, dstname, str(ex)))
        except CTError as err:
            errors.extend(err.errors)
    if errors:
        raise CTError(errors)

def is_locked(filepath):
    """
    Checks if a file is locked by another process.
    """
    import msvcrt
    try:
        fd = os.open(filepath, os.O_APPEND | os.O_EXCL | os.O_RDWR)
    except OSError:
        return True

    try:
        filesize = os.path.getsize(filepath)
        msvcrt.locking(fd, msvcrt.LK_NBLCK, filesize)
        msvcrt.locking(fd, msvcrt.LK_UNLCK, filesize)
        os.close(fd)
        return False
    except (OSError, IOError):
        os.close(fd)
        return True

def create_file_atomic(filename) -> bool:
    """
    Create a lock file in an atomic way, so it is threadsafe.

    Returns True if the file was created by this thread, False if the file existed already.
    """
    try:
        fd = os.open(filename,  os.O_CREAT | os.O_EXCL)
        os.close(fd)
        return True
    except FileExistsError:
        return False
    except IOError as ex:
        if ex.errno == 13:
            return False

    # If we get here, return False anyway       
    return False



import logging
import math
from typing import Tuple

import geopandas as gpd
import pyproj
import shapely.geometry as sh_geom
import shapely.ops as sh_ops

#-------------------------------------------------------------
# First define/init some general variables/constants
#-------------------------------------------------------------
# Get a logger...
logger = logging.getLogger(__name__)
#logger.setLevel(logging.DEBUG)

#-------------------------------------------------------------
# The real work
#-------------------------------------------------------------

def create_grid(
        total_bounds: Tuple[float, float, float, float],
        nb_columns: int,
        nb_rows: int,
        crs: pyproj.CRS) -> gpd.GeoDataFrame:

    xmin, ymin, xmax, ymax = total_bounds
    width = (xmax-xmin)/nb_columns
    height = (ymax-ymin)/nb_rows

    return create_grid3(total_bounds=total_bounds, width=width, height=height, crs=crs)

def create_grid3(
        total_bounds: Tuple[float, float, float, float],
        width: float,
        height: float,
        crs: pyproj.CRS) -> gpd.GeoDataFrame:

    xmin, ymin, xmax, ymax = total_bounds
    rows = int(math.ceil((ymax-ymin) / height))
    cols = int(math.ceil((xmax-xmin) / width))
     
    polygons = []
    cell_left = xmin
    cell_right = xmin + width
    for _ in range(cols+1):
        if cell_left > xmax:
            break
        cell_top = ymin + height
        cell_bottom = ymin
        for _ in range(rows+1):
            if cell_bottom > ymax:
                break
            polygons.append(sh_ops.Polygon([(cell_left, cell_top), (cell_right, cell_top), (cell_right, cell_bottom), (cell_left, cell_bottom)])) 
            cell_top += height
            cell_bottom += height
            
        cell_left += width
        cell_right += width
        
    return gpd.GeoDataFrame({'geometry': polygons}, crs=crs)

def create_grid2(
        total_bounds: Tuple[float, float, float, float], 
        nb_squarish_tiles: int,
        crs: pyproj.CRS) -> gpd.GeoDataFrame:
    """
    Creates a grid and tries to approximate the number of cells asked as
    good as possible with grid cells that as close to square as possible.

    Args:
        total_bounds (Tuple[float, float, float, float]): bounds of the grid to be created
        nb_squarish_cells (int): about the number of cells wanted

    Returns:
        gpd.GeoDataFrame: geodataframe with the grid
    """
    # If more cells asked, calculate optimal number
    xmin, ymin, xmax, ymax = total_bounds
    total_width = xmax-xmin
    total_height = ymax-ymin

    columns_vs_rows = total_width/total_height
    nb_rows = round(math.sqrt(nb_squarish_tiles/columns_vs_rows))

    # Evade having too many cells if few cells are asked...
    if nb_rows > nb_squarish_tiles:
        nb_rows = nb_squarish_tiles
    nb_columns = round(nb_squarish_tiles/nb_rows)
    
    # Now we know everything to create the grid
    return create_grid(
        total_bounds=total_bounds,
        nb_columns=nb_columns,
        nb_rows=nb_rows,
        crs=crs)

def split_tiles(
        input_tiles: gpd.GeoDataFrame,
        nb_tiles_wanted: int) -> gpd.GeoDataFrame:

    nb_tiles = len(input_tiles)
    if nb_tiles >= nb_tiles_wanted:
        return input_tiles
    
    nb_tiles_ratio_target = nb_tiles_wanted / nb_tiles

    # Loop over all tiles in the grid
    result_tiles = []
    for tile in input_tiles.itertuples():

        # For this tile, as long as the curr_nb_tiles_ratio_todo is not 1, keep splitting 
        curr_nb_tiles_ratio_todo = nb_tiles_ratio_target
        curr_tiles_being_split = [tile.geometry]
        while round(curr_nb_tiles_ratio_todo) > 1:

            # Check in how many parts the tiles are split in this iteration
            divisor = 0 
            if round(curr_nb_tiles_ratio_todo) == 3:
                divisor = 3
            else:
                divisor = 2
            curr_nb_tiles_ratio_todo /= divisor

            # Split all current tiles
            tmp_tiles_after_split = []
            for tile_to_split in curr_tiles_being_split:
                xmin, ymin, xmax, ymax = tile_to_split.bounds
                width = abs(xmax-xmin)
                height = abs(ymax-ymin)

                # Split in 2 or 3...
                if divisor == 3:
                    if width > height:
                        split_line = sh_geom.LineString([
                                (xmin+width/3, ymin-1), (xmin+width/3, ymax+1),
                                (xmin+2*width/3, ymax+1), (xmin+2*width/3, ymin-1)])
                    else:
                        split_line = sh_geom.LineString([
                                (xmin-1, ymin+height/3), (xmax+1, ymin+height/3),
                                (xmax+1, ymin+2*height/3), (xmin-1, ymin+2*height/3)])
                else:
                    if width > height:
                        split_line = sh_geom.LineString([(xmin+width/2, ymin-1), (xmin+width/2, ymax+1)])
                    else:
                        split_line = sh_geom.LineString([(xmin-1, ymin+height/2), (xmax+1, ymin+height/2)])
                tmp_tiles_after_split.extend(sh_ops.split(tile_to_split, split_line))
            curr_tiles_being_split = tmp_tiles_after_split
        result_tiles.extend(curr_tiles_being_split)
    
    # We should be ready...
    return gpd.GeoDataFrame(geometry=result_tiles, crs=input_tiles.crs)

def extract_polygons_from_list(
        in_geom: sh_geom.base.BaseGeometry) -> list:
    """
    Extracts all polygons from the input geom and returns them as a list.
    """
    # Extract the polygons from the multipolygon, but store them as multipolygons anyway
    geoms = []
    if in_geom.geom_type == 'MultiPolygon':
        geoms = list(in_geom)
    elif in_geom.geom_type == 'Polygon':
        geoms.append(in_geom)
    elif in_geom.geom_type == 'GeometryCollection':
        for geom in in_geom:
            if geom.geom_type == 'MultiPolygon':
                geoms.append(list(geom))
            elif geom.geom_type == 'Polygon':
                geoms.append(geom)
            else:
                logger.debug(f"Found {geom.geom_type}, ignore!")
    else:
        raise IOError(f"in_geom is of an unsupported type: {in_geom.geom_type}")
    
    return geoms

def extract_polygons_from_gdf(
        in_gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:

    # Extract only polygons
    poly_gdf = in_gdf.loc[(in_gdf.geometry.geom_type == 'Polygon')].copy()
    multipoly_gdf = in_gdf.loc[(in_gdf.geometry.geom_type == 'MultiPolygon')].copy()
    collection_gdf = in_gdf.loc[(in_gdf.geometry.geom_type == 'GeometryCollection')].copy()
    collection_polys_gdf = None
    
    if len(collection_gdf) > 0:
        collection_polygons = []
        for collection_geom in collection_gdf.geometry:
            collection_polygons.extend(extract_polygons_from_list(collection_geom))
        if len(collection_polygons) > 0:
            collection_polys_gdf = gpd.GeoDataFrame(geometry=collection_polygons, crs=in_gdf.crs)

    # Only keep the polygons...
    ret_gdf = poly_gdf
    if len(multipoly_gdf) > 0:
        ret_gdf = ret_gdf.append(multipoly_gdf.explode(), ignore_index=True)
    if collection_polys_gdf is not None:
        ret_gdf = ret_gdf.append(collection_polys_gdf, ignore_index=True)
    
    return ret_gdf

def polygons_to_lines(input_gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    cardsheets_lines = []
    for cardsheet_poly in input_gdf.itertuples():
        cardsheet_boundary = cardsheet_poly.geometry.boundary
        if cardsheet_boundary.type == 'MultiLineString':
            for line in cardsheet_boundary:
                cardsheets_lines.append(line)
        else:
            cardsheets_lines.append(cardsheet_boundary)

    cardsheets_lines_gdf = gpd.GeoDataFrame(geometry=cardsheets_lines, crs=input_gdf.crs)

    return cardsheets_lines_gdf




from pathlib import Path
import sys
sys.path.insert(0, str(Path(__file__).resolve().parent / '..'))

from geofileops import geofile

def get_testdata_dir() -> Path:
    return Path(__file__).resolve().parent / 'data'

def test_listlayers():
    # Test shapefile
    srcpath = get_testdata_dir() / 'parcels.shp'
    layers = geofile.listlayers(srcpath)
    assert layers[0] == 'parcels'
    
    # Test geopackage
    srcpath = get_testdata_dir() / 'parcels.gpkg'
    layers = geofile.listlayers(srcpath)
    assert layers[0] == 'parcels'

def test_getlayerinfo():
    # Test shapefile
    srcpath = get_testdata_dir() / 'parcels.shp'
    layerinfo = geofile.getlayerinfo(srcpath)
    assert layerinfo.featurecount == 46
    assert layerinfo.geometrycolumn == 'geometry' 

    # Test geopackage
    srcpath = get_testdata_dir() / 'parcels.gpkg'
    layerinfo = geofile.getlayerinfo(srcpath, 'parcels')
    assert layerinfo.featurecount == 46
    assert layerinfo.geometrycolumn == 'geom' 

def test_get_only_layer():
    srcpath = get_testdata_dir() / 'parcels.shp'
    layer = geofile.get_only_layer(srcpath)
    assert layer == 'parcels'

def test_get_default_layer():
    srcpath = get_testdata_dir() / 'parcels.shp'
    layer = geofile.get_default_layer(srcpath)
    assert layer == 'parcels'

def test_spatial_index_gpkg(tmpdir):
    # First copy test file to tmpdir
    src = get_testdata_dir() / 'parcels.gpkg'
    tmppath = Path(tmpdir) / 'parcels.gpkg'
    geofile.copy(src, tmppath)

    # Check if spatial index present
    has_spatial_index = geofile.has_spatial_index(
        path=tmppath, layer='parcels')
    assert has_spatial_index is True

    # Remove spatial index
    geofile.remove_spatial_index(path=tmppath, layer='parcels')
    has_spatial_index = geofile.has_spatial_index(
        path=tmppath, layer='parcels')
    assert has_spatial_index is False

    # Create spatial index
    geofile.create_spatial_index(path=tmppath, layer='parcels')
    has_spatial_index = geofile.has_spatial_index(
        path=tmppath, layer='parcels')
    assert has_spatial_index is True

def test_spatial_index_shp(tmpdir):
    # First copy test file to tmpdir
    src = get_testdata_dir() / 'parcels.shp'
    tmppath = Path(tmpdir) / 'parcels.shp'
    geofile.copy(src, tmppath)

    # Check if spatial index present
    has_spatial_index = geofile.has_spatial_index(path=tmppath)
    assert has_spatial_index is True

    # Remove spatial index
    geofile.remove_spatial_index(path=tmppath)
    has_spatial_index = geofile.has_spatial_index(path=tmppath)
    assert has_spatial_index is False

    # Create spatial index
    geofile.create_spatial_index(path=tmppath)
    has_spatial_index = geofile.has_spatial_index(path=tmppath)
    assert has_spatial_index is True

def test_rename_layer(tmpdir):
    # First copy test file to tmpdir
    src = get_testdata_dir() / 'parcels.gpkg'
    tmppath = Path(tmpdir) / 'parcels.gpkg'
    geofile.copy(src, tmppath)

    # Now test rename layer
    geofile.rename_layer(tmppath, layer='parcels', new_layer='parcels_renamed')
    layernames_renamed = geofile.listlayers(path=tmppath)
    assert layernames_renamed[0] == 'parcels_renamed'

def test_add_column(tmpdir):
    # First copy test file to tmpdir
    # Now add area column
    src = get_testdata_dir() / 'parcels.gpkg'
    tmppath = Path(tmpdir) / 'parcels.gpkg'
    geofile.copy(src, tmppath)

    # The area column shouldn't be in the test file yet
    layerinfo = geofile.getlayerinfo(path=tmppath, layer='parcels')
    assert 'area' not in layerinfo.columns
        
    # Add area column
    try: 
        import os
        os.environ['GDAL_BIN'] = r"X:\GIS\Software\_Progs\OSGeo4W64_2020-05-29\bin"
        geofile.add_column(tmppath, layer='parcels', name='area', type='real', expression='ST_area(geom)')
    finally:
        del os.environ['GDAL_BIN']
        
    layerinfo = geofile.getlayerinfo(path=tmppath, layer='parcels')
    assert 'area' in layerinfo.columns
    
    gdf = geofile.read_file(tmppath)
    assert round(gdf['area'][0], 1) == round(gdf['OPPERVL'][0], 1)

def test_read_file():
    # Test shapefile
    srcpath = get_testdata_dir() / 'parcels.shp'
    read_gdf = geofile.read_file(srcpath)
    assert len(read_gdf) == 46

    # Test geopackage
    srcpath = get_testdata_dir() / 'parcels.gpkg'
    read_gdf = geofile.read_file(srcpath)
    assert len(read_gdf) == 46

def test_to_file_shp(tmpdir):
    # Read test file and write to tmpdir
    srcpath = get_testdata_dir() / 'parcels.shp'
    read_gdf = geofile.read_file(srcpath)
    tmppath = Path(tmpdir) / 'parcels.shp'
    geofile.to_file(read_gdf, tmppath)
    tmp_gdf = geofile.read_file(tmppath)
    
    assert len(read_gdf) == len(tmp_gdf)

def test_to_file_gpkg(tmpdir):
    # Read test file and write to tmpdir
    srcpath = get_testdata_dir() / 'parcels.gpkg'
    read_gdf = geofile.read_file(srcpath)
    tmppath = Path(tmpdir) / 'parcels.gpkg'
    geofile.to_file(read_gdf, tmppath)
    tmp_gdf = geofile.read_file(tmppath)
    
    assert len(read_gdf) == len(tmp_gdf)

def test_get_crs():
    # Test shapefile
    srcpath = get_testdata_dir() / 'parcels.shp'
    crs = geofile.get_crs(srcpath)
    assert crs.to_epsg() == 31370

    # Test geopackage
    srcpath = get_testdata_dir() / 'parcels.gpkg'
    crs = geofile.get_crs(srcpath)
    assert crs.to_epsg() == 31370

def test_is_geofile():
    # Test shapefile
    srcpath = get_testdata_dir() / 'parcels.shp'
    assert geofile.is_geofile(srcpath) == True

    # Test geopackage
    srcpath = get_testdata_dir() / 'parcels.gpkg'
    assert geofile.is_geofile(srcpath) == True

def test_copy(tmpdir):
    # Copy test file to tmpdir
    src = get_testdata_dir() / 'parcels.shp'
    dst = Path(tmpdir) / 'parcels.shp'
    geofile.copy(src, dst)
    assert dst.exists() == True

def test_cmp(tmpdir):
    # Copy test file to tmpdir
    src = get_testdata_dir() / 'parcels.shp'
    dst = Path(tmpdir) / 'parcels.shp'
    geofile.copy(src, dst)

    # Now compare source and dst file
    assert geofile.cmp(src, dst) == True

def test_move(tmpdir):
    # Copy test file to tmpdir
    src = get_testdata_dir() / 'parcels.shp'
    tmp1path = Path(tmpdir) / 'parcels.shp'
    geofile.copy(src, tmp1path)
    assert tmp1path.exists() == True

    # Move (rename actually) and check result
    tmp2path = Path(tmpdir) / 'parcels2.shp'
    geofile.move(tmp1path, tmp2path)
    assert tmp1path.exists() == False
    assert tmp2path.exists() == True

def test_remove(tmpdir):
    # Copy test file to tmpdir
    src = get_testdata_dir() / 'parcels.shp'
    tmppath = Path(tmpdir) / 'parcels.shp'
    geofile.copy(src, tmppath)
    assert tmppath.exists() == True

    # Remove and check result
    geofile.remove(tmppath)
    assert tmppath.exists() == False

def test_get_driver():
    # Test shapefile
    src = get_testdata_dir() / 'parcels.shp'
    assert geofile.get_driver(src) == "ESRI Shapefile"

    # Test geopackage
    src = get_testdata_dir() / 'parcels.gpkg'
    assert geofile.get_driver(src) == "GPKG"

if __name__ == '__main__':
    import tempfile
    
    tmpdir = tempfile.gettempdir()
    test_add_column(tmpdir)




import os
from pathlib import Path
import sys
sys.path.insert(0, str(Path(__file__).resolve().parent / '..'))

from geofileops import geofile
from geofileops.util import geofileops_ogr

class GdalBin():
    def __init__(self, set_gdal_bin: bool, gdal_bin_path: str = None):
        self.set_gdal_bin = set_gdal_bin
        if set_gdal_bin is True:
            if gdal_bin_path is None:
                self.gdal_bin_path = r"X:\GIS\Software\_Progs\OSGeo4W64_2020-05-29\bin"
            else:
                self.gdal_bin_path = gdal_bin_path

    def __enter__(self):
        if self.set_gdal_bin is True:
            import os
            os.environ['GDAL_BIN'] = self.gdal_bin_path

    def __exit__(self, type, value, traceback):
        #Exception handling here
        import os
        if os.environ['GDAL_BIN'] is not None:
            del os.environ['GDAL_BIN']

def get_testdata_dir() -> Path:
    return Path(__file__).resolve().parent / 'data'

def is_gdal_default_ok() -> bool:
    return False

def is_gdal_bin_ok() -> bool:
    return True

def test_select_shp(tmpdir):
    # Select some data from src to tmp file
    input_path = get_testdata_dir() / 'parcels.shp'
    output_path = Path(tmpdir) / 'parcels.shp'
    layerinfo_orig = geofile.getlayerinfo(input_path)
    sql_stmt = f'select {layerinfo_orig.geometrycolumn}, oidn, uidn from "parcels"'
    geofileops_ogr.select(
            input_path=input_path,
            output_path=output_path,
            sql_stmt=sql_stmt,
            sql_dialect='SQLITE')

    # Now check if the tmp file is correctly created
    layerinfo_select = geofile.getlayerinfo(output_path)
    assert layerinfo_orig.featurecount == layerinfo_select.featurecount
    assert 'OIDN' in layerinfo_select.columns
    assert 'UIDN' in layerinfo_select.columns
    assert len(layerinfo_select.columns) == 2

    output_gdf = geofile.read_file(output_path)
    assert output_gdf['geometry'][0] is not None

def test_select_gpkg(tmpdir):
    # Select some data from src to tmp file
    input_path = get_testdata_dir() / 'parcels.gpkg'
    output_path = Path(tmpdir) / 'parcels.gpkg'
    layerinfo_orig = geofile.getlayerinfo(input_path)
    sql_stmt = f'select {layerinfo_orig.geometrycolumn}, oidn, uidn from "parcels"'
    geofileops_ogr.select(
            input_path=input_path,
            output_path=output_path,
            sql_stmt=sql_stmt)

    # Now check if the tmp file is correctly created
    layerinfo_select = geofile.getlayerinfo(output_path)
    assert layerinfo_orig.featurecount == layerinfo_select.featurecount
    assert 'OIDN' in layerinfo_select.columns
    assert 'UIDN' in layerinfo_select.columns
    assert len(layerinfo_select.columns) == 2

def test_isvalid_gpkg(tmpdir):
    # Buffer to test dir
    input_path = get_testdata_dir() / 'parcels.gpkg'
    output_path = Path(tmpdir) / 'parcels.gpkg'

    # Try with and without gdal_bin
    basetest_isvalid(input_path, output_path, set_gdal_bin=True, ok_expected=is_gdal_bin_ok())
    basetest_isvalid(input_path, output_path, set_gdal_bin=False, ok_expected=is_gdal_default_ok())

def test_isvalid_shp(tmpdir):
    # Select some data from src to tmp file
    input_path = get_testdata_dir() / 'parcels.shp'
    output_path = Path(tmpdir) / 'parcels.gpkg'
    
    # Try with and without gdal_bin
    basetest_isvalid(input_path, output_path, set_gdal_bin=True, ok_expected=is_gdal_bin_ok())
    basetest_isvalid(input_path, output_path, set_gdal_bin=False, ok_expected=is_gdal_default_ok())
    
def basetest_isvalid(
        input_path: Path, 
        output_path: Path, 
        set_gdal_bin: bool, 
        ok_expected: bool):
    
    # Do operation
    try:
        with GdalBin(set_gdal_bin):
            assert geofileops_ogr.isvalid(input_path=input_path, output_path=output_path) == True
        test_ok = True
    except:
        test_ok = False
    assert test_ok is ok_expected, "Without gdal_bin set to an osgeo installation, it is 'normal' this fails"

    # If it is expected not to be OK, don't do other checks
    if ok_expected is False:
        return

    '''
    # Now check if the tmp file is correctly created
    assert output_path.exists() == True
    layerinfo_select = geofile.getlayerinfo(output_path)
    assert layerinfo_orig.featurecount == layerinfo_select.featurecount
    assert (len(layerinfo_orig.columns)+3) == len(layerinfo_select.columns)

    output_gdf = geofile.read_file(output_path)
    print(output_gdf)
    assert output_gdf['geom'][0] is None
    assert output_gdf['isvalid'][0] == 1
    assert output_gdf['isvalidreason'][0] == 'Valid Geometry'
    '''

def test_convexhull_gpkg(tmpdir):
    # Execute to test dir
    input_path = get_testdata_dir() / 'parcels.gpkg'
    output_path = Path(tmpdir) / 'parcels.gpkg'
    
    # Without gdal_bin set, this fails at the moment
    basetest_convexhull(input_path, output_path, set_gdal_bin=True, ok_expected=is_gdal_bin_ok())
    basetest_convexhull(input_path, output_path, set_gdal_bin=False, ok_expected=is_gdal_default_ok())

def test_convexhull_shp(tmpdir):
    # Select some data from src to tmp file
    input_path = get_testdata_dir() / 'parcels.shp'
    output_path = Path(tmpdir) / 'parcels.shp'

    # Without gdal_bin set, this fails at the moment
    basetest_convexhull(input_path, output_path, set_gdal_bin=True, ok_expected=is_gdal_bin_ok())
    basetest_convexhull(input_path, output_path, set_gdal_bin=False, ok_expected=is_gdal_default_ok())

def basetest_convexhull(
        input_path: Path, 
        output_path: Path, 
        set_gdal_bin: bool, 
        ok_expected: bool):
    
    # Do operation  
    try:
        with GdalBin(set_gdal_bin):
            geofileops_ogr.convexhull(input_path=input_path, output_path=output_path)
        test_ok = True
    except:
        test_ok = False
    assert test_ok is ok_expected, "Without gdal_bin set to an osgeo installation, it is 'normal' this fails"

    # If it is expected not to be OK, don't do other checks
    if ok_expected is False:
        return

    # Now check if the tmp file is correctly created
    assert output_path.exists() == True
    layerinfo_orig = geofile.getlayerinfo(input_path)
    layerinfo_select = geofile.getlayerinfo(output_path)
    assert layerinfo_orig.featurecount == layerinfo_select.featurecount
    assert len(layerinfo_orig.columns) == len(layerinfo_select.columns)

    output_gdf = geofile.read_file(output_path)
    assert output_gdf['geometry'][0] is not None

def test_buffer_gpkg(tmpdir):
    # Buffer to test dir
    input_path = get_testdata_dir() / 'parcels.gpkg'
    output_path = Path(tmpdir) / 'parcels.gpkg'

    # Without gdal_bin set, this fails at the moment
    basetest_buffer(input_path, output_path, set_gdal_bin=True, ok_expected=is_gdal_bin_ok())
    basetest_buffer(input_path, output_path, set_gdal_bin=False, ok_expected=is_gdal_default_ok())
        
def test_buffer_shp(tmpdir):
    # Buffer to test dir
    input_path = get_testdata_dir() / 'parcels.shp'
    output_path = Path(tmpdir) / 'parcels.shp'

    # Without gdal_bin set, this fails at the moment
    basetest_buffer(input_path, output_path, set_gdal_bin=True, ok_expected=is_gdal_bin_ok())
    basetest_buffer(input_path, output_path, set_gdal_bin=False, ok_expected=is_gdal_default_ok())

def basetest_buffer(
        input_path: Path, 
        output_path: Path, 
        set_gdal_bin: bool, 
        ok_expected: bool):
    
    # Do operation
    try:
        with GdalBin(set_gdal_bin):
            geofileops_ogr.buffer(input_path=input_path, output_path=output_path, distance=1)
        test_ok = True
    except:
        test_ok = False
    assert test_ok is ok_expected, "Without gdal_bin set to an osgeo installation, it is 'normal' this fails"

    # If it is expected not to be OK, don't do other checks
    if ok_expected is False:
        return

    # Now check if the tmp file is correctly created
    assert output_path.exists() == True
    layerinfo_orig = geofile.getlayerinfo(input_path)
    layerinfo_select = geofile.getlayerinfo(input_path)
    assert layerinfo_orig.featurecount == layerinfo_select.featurecount
    assert len(layerinfo_orig.columns) == len(layerinfo_select.columns)

    output_gdf = geofile.read_file(output_path)
    assert output_gdf['geometry'][0] is not None
    geofile.remove(output_path)

def test_makevalid_gpkg(tmpdir):
    # makevalid to test dir
    input_path = get_testdata_dir() / 'invalid_geometries.gpkg'
    output_path = Path(tmpdir) / input_path.name

    # Without gdal_bin set, this fails at the moment
    basetest_makevalid(input_path, output_path, set_gdal_bin=True, ok_expected=is_gdal_bin_ok())
    basetest_makevalid(input_path, output_path, set_gdal_bin=False, ok_expected=is_gdal_default_ok())
           
def basetest_makevalid(
        input_path: Path, 
        output_path: Path, 
        set_gdal_bin: bool, 
        ok_expected: bool):

    # Do operation
    try:
        with GdalBin(set_gdal_bin):
            geofileops_ogr.makevalid(input_path=input_path, output_path=output_path)
        test_ok = True
    except:
        test_ok = False
    assert test_ok is ok_expected, "Without gdal_bin set to an osgeo installation, it is 'normal' this fails"

    # If it is expected not to be OK, don't do other checks
    if ok_expected is False:
        return

    # Now check if the output file is correctly created
    assert output_path.exists() == True
    layerinfo_orig = geofile.getlayerinfo(input_path)
    layerinfo_output = geofile.getlayerinfo(output_path)
    assert layerinfo_orig.featurecount == layerinfo_output.featurecount
    assert len(layerinfo_orig.columns) == len(layerinfo_output.columns)

    output_gdf = geofile.read_file(output_path)
    assert output_gdf['geometry'][0] is not None

    # Make sure the input file was not valid
    output_isvalid_path = output_path.parent / f"{output_path.stem}_isvalid{output_path.suffix}"
    with GdalBin(set_gdal_bin):
        isvalid = geofileops_ogr.isvalid(input_path=input_path, output_path=output_isvalid_path)
    assert isvalid is False, "Input file should contain invalid features"

    # Check if the result file is valid
    output_new_isvalid_path = output_path.parent / f"{output_path.stem}_new_isvalid{output_path.suffix}"
    with GdalBin(set_gdal_bin):
        isvalid = geofileops_ogr.isvalid(input_path=output_path, output_path=output_new_isvalid_path)
    assert isvalid == True, "Output file shouldn't contain invalid features"

def test_simplify_gpkg(tmpdir):
    # Simplify to test dir
    input_path = get_testdata_dir() / 'parcels.gpkg'
    output_path = Path(tmpdir) / input_path.name

    # Without gdal_bin set, this fails with libspatialite 4.3
    basetest_simplify(input_path, output_path, set_gdal_bin=True, ok_expected=is_gdal_bin_ok())
    basetest_simplify(input_path, output_path, set_gdal_bin=False, ok_expected=is_gdal_default_ok())
        
def test_simplify_shp(tmpdir):
    # Simplify to test dir
    input_path = get_testdata_dir() / 'parcels.shp'
    output_path = Path(tmpdir) / 'parcels.shp'

    # Without gdal_bin set, this fails with libspatialite 4.3
    basetest_simplify(input_path, output_path, set_gdal_bin=True, ok_expected=is_gdal_bin_ok())
    basetest_simplify(input_path, output_path, set_gdal_bin=False, ok_expected=is_gdal_default_ok())
    
def basetest_simplify(
        input_path: Path, 
        output_path: Path,
        set_gdal_bin: bool, 
        ok_expected: bool):

    # Do operation
    try:
        with GdalBin(set_gdal_bin):
            geofileops_ogr.simplify(
                    input_path=input_path, output_path=output_path,
                    tolerance=5)
        test_ok = True
    except:
        test_ok = False
    assert test_ok is ok_expected, "Without gdal_bin set to an osgeo installation, it is 'normal' this fails"

    # If it is expected not to be OK, don't do other checks
    if ok_expected is False:
        return

    # Now check if the output file is correctly created
    assert output_path.exists() == True
    layerinfo_orig = geofile.getlayerinfo(input_path)
    layerinfo_output = geofile.getlayerinfo(output_path)
    assert layerinfo_orig.featurecount == layerinfo_output.featurecount
    assert len(layerinfo_orig.columns) == len(layerinfo_output.columns)

    output_gdf = geofile.read_file(output_path)
    assert output_gdf['geometry'][0] is not None

def test_erase_gpkg(tmpdir):
    # Erase to test dir
    input_path = get_testdata_dir() / 'parcels.gpkg'
    erase_path = get_testdata_dir() / 'zones.gpkg'
    output_path = Path(tmpdir) / 'parcels.gpkg'

    # Without gdal_bin set, this fails with libspatialite 4.3
    basetest_erase(input_path, erase_path, output_path, set_gdal_bin=True, ok_expected=is_gdal_bin_ok())
    basetest_erase(input_path, erase_path, output_path, set_gdal_bin=False, ok_expected=is_gdal_default_ok())

def test_erase_shp(tmpdir):
    # Buffer to test dir
    input_path = get_testdata_dir() / 'parcels.shp'
    erase_path = get_testdata_dir() / 'zones.gpkg'
    output_path = Path(tmpdir) / 'parcels.gpkg'

    # Without gdal_bin set, this fails with libspatialite 4.3
    basetest_erase(input_path, erase_path, output_path, set_gdal_bin=True, ok_expected=is_gdal_bin_ok())
    basetest_erase(input_path, erase_path, output_path, set_gdal_bin=False, ok_expected=is_gdal_default_ok())

def basetest_erase(
        input_path: Path,
        erase_path: Path, 
        output_path: Path, 
        set_gdal_bin: bool, 
        ok_expected: bool):

    # Do operation
    try:
        with GdalBin(set_gdal_bin):
            geofileops_ogr.erase(
                    input_path=input_path, erase_path=erase_path,
                    output_path=output_path)
        test_ok = True
    except:
        test_ok = False
    assert test_ok is ok_expected, "Without gdal_bin set to an osgeo installation, it is 'normal' this fails"

    # If it is expected not to be OK, don't do other checks
    if ok_expected is False:
        return

    # Now check if the tmp file is correctly created
    assert output_path.exists() == True
    layerinfo_orig = geofile.getlayerinfo(input_path)
    layerinfo_select = geofile.getlayerinfo(input_path)
    assert layerinfo_orig.featurecount == layerinfo_select.featurecount
    assert len(layerinfo_orig.columns) == len(layerinfo_select.columns)

    output_gdf = geofile.read_file(output_path)
    assert output_gdf['geometry'][0] is not None

def test_export_by_location_gpkg(tmpdir):
    # Export to test dir
    input_to_select_from_path = get_testdata_dir() / 'parcels.gpkg'
    input_to_compare_with_path = get_testdata_dir() / 'zones.gpkg'
    output_path = Path(tmpdir) / 'parcels.gpkg'

    # Without gdal_bin set, this fails with libspatialite 4.3
    basetest_export_by_location(
            input_to_select_from_path, input_to_compare_with_path, output_path, 
            set_gdal_bin=True, ok_expected=is_gdal_bin_ok())
    basetest_export_by_location(
            input_to_select_from_path, input_to_compare_with_path, output_path, 
            set_gdal_bin=False, ok_expected=is_gdal_default_ok())
        
def test_export_by_location_shp(tmpdir):
    # Export to test dir
    input_to_select_from_path = get_testdata_dir() / 'parcels.shp'
    input_to_compare_with_path = get_testdata_dir() / 'zones.gpkg'
    output_path = Path(tmpdir) / 'parcels.gpkg'

    # Without gdal_bin set, this fails at the moment
    basetest_export_by_location(
            input_to_select_from_path, input_to_compare_with_path, output_path, 
            set_gdal_bin=True, ok_expected=is_gdal_bin_ok())
    basetest_export_by_location(
            input_to_select_from_path, input_to_compare_with_path, output_path, 
            set_gdal_bin=False, ok_expected=is_gdal_default_ok())

def basetest_export_by_location(
        input_to_select_from_path: Path, 
        input_to_compare_with_path: Path, 
        output_path: Path, 
        set_gdal_bin: bool, 
        ok_expected: bool):

    # Do operation
    try:
        with GdalBin(set_gdal_bin):
            geofileops_ogr.export_by_location(
                    input_to_select_from_path=input_to_select_from_path,
                    input_to_compare_with_path=input_to_compare_with_path,
                    output_path=output_path)
        test_ok = True
    except:
        test_ok = False
    assert test_ok is ok_expected, "Without gdal_bin set to an osgeo installation, it is 'normal' this fails"

    # If it is expected not to be OK, don't do other checks
    if ok_expected is False:
        return

    # Now check if the tmp file is correctly created
    assert output_path.exists() == True
    layerinfo_orig = geofile.getlayerinfo(input_to_select_from_path)
    layerinfo_select = geofile.getlayerinfo(input_to_select_from_path)
    assert layerinfo_orig.featurecount == layerinfo_select.featurecount
    assert len(layerinfo_orig.columns) == len(layerinfo_select.columns)

    output_gdf = geofile.read_file(output_path)
    assert output_gdf['geometry'][0] is not None

def test_export_by_distance_gpkg(tmpdir):
        # Export to test dir
    input_to_select_from_path = get_testdata_dir() / 'parcels.gpkg'
    input_to_compare_with_path = get_testdata_dir() / 'zones.gpkg'
    output_path = Path(tmpdir) / 'parcels.gpkg'
    
    # Without gdal_bin set, this fails at the moment
    basetest_export_by_location(
            input_to_select_from_path, input_to_compare_with_path, output_path, 
            set_gdal_bin=True, ok_expected=is_gdal_bin_ok())
    basetest_export_by_location(
            input_to_select_from_path, input_to_compare_with_path, output_path, 
            set_gdal_bin=False, ok_expected=is_gdal_default_ok())
    
def test_export_by_distance_shp(tmpdir):
    # Export to test dir
    input_to_select_from_path = get_testdata_dir() / 'parcels.shp'
    input_to_compare_with_path = get_testdata_dir() / 'zones.gpkg'
    output_path = Path(tmpdir) / 'parcels.gpkg'

    # Without gdal_bin set, this fails at the moment
    basetest_export_by_location(
            input_to_select_from_path, input_to_compare_with_path, output_path, 
            set_gdal_bin=True, ok_expected=is_gdal_bin_ok())
    basetest_export_by_location(
            input_to_select_from_path, input_to_compare_with_path, output_path, 
            set_gdal_bin=False, ok_expected=is_gdal_default_ok())

def basetest_export_by_distance(
        input_to_select_from_path: Path, 
        input_to_compare_with_path: Path, 
        output_path: Path,
        set_gdal_bin: bool, 
        ok_expected: bool):

    # Do operation
    try:
        with GdalBin(set_gdal_bin):
            geofileops_ogr.export_by_distance(
                    input_to_select_from_path=input_to_select_from_path,
                    input_to_compare_with_path=input_to_compare_with_path,
                    max_distance=10,
                    output_path=output_path)
        test_ok = True
    except:
        test_ok = False
    assert test_ok is ok_expected, "Without gdal_bin set to an osgeo installation, it is 'normal' this fails"

    # If it is expected not to be OK, don't do other checks
    if ok_expected is False:
        return

    # Now check if the tmp file is correctly created
    assert output_path.exists() == True
    layerinfo_orig = geofile.getlayerinfo(input_to_select_from_path)
    layerinfo_select = geofile.getlayerinfo(input_to_select_from_path)
    assert layerinfo_orig.featurecount == layerinfo_select.featurecount
    assert len(layerinfo_orig.columns) == len(layerinfo_select.columns)

    output_gdf = geofile.read_file(output_path)
    assert output_gdf['geometry'][0] is not None

def test_intersect_gpkg(tmpdir):
    # Export to test dir
    input1_path = get_testdata_dir() / 'parcels.gpkg'
    input2_path = get_testdata_dir() / 'zones.gpkg'
    output_path = Path(tmpdir) / 'parcels_intersect_zones.gpkg'

    # Without gdal_bin set, this fails at the moment
    basetest_intersect(input1_path, input2_path, output_path, set_gdal_bin=False, ok_expected=is_gdal_default_ok())
    basetest_intersect(input1_path, input2_path, output_path, set_gdal_bin=True, ok_expected=is_gdal_bin_ok())
    
def test_intersect_shp(tmpdir):
    # Export to test dir
    input1_path = get_testdata_dir() / 'parcels.shp'
    input2_path = get_testdata_dir() / 'zones.gpkg'
    output_path = Path(tmpdir) / 'parcels_intersect_zones.gpkg'
    
    # Without gdal_bin set, this fails at the moment
    basetest_intersect(input1_path, input2_path, output_path, set_gdal_bin=False, ok_expected=is_gdal_default_ok())
    basetest_intersect(input1_path, input2_path, output_path, set_gdal_bin=True, ok_expected=is_gdal_bin_ok())
    
def basetest_intersect(
        input1_path: Path, 
        input2_path: Path, 
        output_path: Path, 
        set_gdal_bin: bool, 
        ok_expected: bool):

    # Do operation
    try:
        with GdalBin(set_gdal_bin):
            geofileops_ogr.intersect(
                    input1_path=input1_path,
                    input2_path=input2_path,
                    output_path=output_path,
                    verbose=True)
        test_ok = True
    except:
        test_ok = False
    assert test_ok is ok_expected, "Without gdal_bin set to an osgeo installation, it is 'normal' this fails"

    # If it is expected not to be OK, don't do other checks
    if ok_expected is False:
        return

    # Now check if the tmp file is correctly created
    assert output_path.exists() == True
    layerinfo_input1 = geofile.getlayerinfo(input1_path)
    layerinfo_input2 = geofile.getlayerinfo(input2_path)
    layerinfo_select = geofile.getlayerinfo(output_path)
    assert layerinfo_select.featurecount == 28
    assert (len(layerinfo_input1.columns) + len(layerinfo_input2.columns)) == len(layerinfo_select.columns)

    output_gdf = geofile.read_file(output_path)
    assert output_gdf['geometry'][0] is not None

def test_join_by_location_gpkg(tmpdir):
    # Export to test dir
    input1_path = get_testdata_dir() / 'parcels.gpkg'
    input2_path = get_testdata_dir() / 'zones.gpkg'
    output_path = Path(tmpdir) / 'parcels.gpkg'
    
    # Without gdal_bin set, this fails at the moment
    basetest_join_by_location(input1_path, input2_path, output_path, set_gdal_bin=False, ok_expected=is_gdal_default_ok())
    basetest_join_by_location(input1_path, input2_path, output_path, set_gdal_bin=True, ok_expected=is_gdal_bin_ok())
    
def test_join_by_location_shp(tmpdir):
    # Export to test dir
    input1_path = get_testdata_dir() / 'parcels.shp'
    input2_path = get_testdata_dir() / 'zones.gpkg'
    output_path = Path(tmpdir) / 'parcels.gpkg'
    
    # Without gdal_bin set, this fails at the moment
    basetest_join_by_location(input1_path, input2_path, output_path, set_gdal_bin=False, ok_expected=is_gdal_default_ok())
    basetest_join_by_location(input1_path, input2_path, output_path, set_gdal_bin=True, ok_expected=is_gdal_bin_ok())
    
def basetest_join_by_location(
        input1_path: Path, input2_path: Path, output_path: Path, 
        set_gdal_bin: bool, ok_expected: bool):
        
    ### Test 1: inner join, intersect
    try:
        with GdalBin(set_gdal_bin):
            geofileops_ogr.join_by_location(
                    input1_path=input1_path,
                    input2_path=input2_path,
                    output_path=output_path,
                    force=True)
        test_ok = True
    except:
        test_ok = False
    assert test_ok is ok_expected, "Without gdal_bin set to an osgeo installation, it is 'normal' this fails"

    # If it is expected not to be OK, don't do other checks
    if ok_expected is False:
        return

    # Now check if the output file is correctly created
    assert output_path.exists() == True
    layerinfo_input1 = geofile.getlayerinfo(input1_path)
    layerinfo_input2 = geofile.getlayerinfo(input2_path)
    layerinfo_result = geofile.getlayerinfo(output_path)
    assert layerinfo_result.featurecount == 4
    assert (len(layerinfo_input1.columns) + len(layerinfo_input2.columns) + 1) == len(layerinfo_result.columns)

    output_gdf = geofile.read_file(output_path)
    assert output_gdf['geometry'][0] is not None

    ### Test 2: left outer join, intersect
    try:
        with GdalBin(set_gdal_bin):
            geofileops_ogr.join_by_location(
                    input1_path=input1_path,
                    input2_path=input2_path,
                    output_path=output_path,
                    discard_nonmatching=False,
                    force=True)
        test_ok = True
    except:
        test_ok = False
    assert test_ok is ok_expected, "Without gdal_bin set to an osgeo installation, it is 'normal' this fails"

    # If it is expected not to be OK, don't do other checks
    if ok_expected is False:
        return

    # Now check if the output file is correctly created
    assert output_path.exists() == True
    layerinfo_result = geofile.getlayerinfo(output_path)
    assert layerinfo_result.featurecount == 24, f"Featurecount is {layerinfo_result.featurecount}, expected 48"
    assert (len(layerinfo_input1.columns) + len(layerinfo_input2.columns) + 1) == len(layerinfo_result.columns)

    output_gdf = geofile.read_file(output_path)
    assert output_gdf['geometry'][0] is not None

if __name__ == '__main__':
    import tempfile
    import shutil
    tmpdir = Path(tempfile.gettempdir()) / 'test_geofileops_ogr_ogr'
    if tmpdir.exists():
        shutil.rmtree(tmpdir)
    
    # Single layer operations
    test_buffer_gpkg(tmpdir)
    #test_makevalid_gpkg(tmpdir)
    #test_erase_shp(tmpdir)
    #test_isvalid_shp(tmpdir)
    #test_isvalid_gpkg(tmpdir)
    #test_convexhull_shp(tmpdir)
    #test_convexhull_gpkg(tmpdir)
    #test_select_geos_version(tmpdir)

    # Two layer operations
    #test_intersect_gpkg(tmpdir)
    #test_export_by_distance_shp(tmpdir)
    #test_join_by_location_gpkg(tmpdir)
    



import datetime
import logging
import math
import multiprocessing
import os
from typing import NamedTuple, Optional

import psutil

################################################################################
# Some init
################################################################################

logger = logging.getLogger(__name__)

################################################################################
# The real work
################################################################################

def initgdal():
    # At least on windows, the init of this doensn't seem to work properly... 
    # TODO: should be solved somewhere else?
    if os.environ['PROJ_LIB'] is None:
        os.environ['PROJ_LIB'] = r"C:\Tools\miniconda3\envs\orthoseg\Library\share\proj"
        logger.warn(f"PROJ_LIB environment variable was not set, set to {os.environ['PROJ_LIB']}")
    if os.environ['GDAL_DATA'] is None:
        os.environ['GDAL_DATA'] = r"C:\Tools\miniconda3\envs\orthoseg\Library\share\gdal"
        logger.warn(f"GDAL_DATA environment variable was not set, set to {os.environ['GDAL_DATA']}")

def report_progress(
        start_time: datetime.datetime,
        nb_done: int,
        nb_todo: int,
        operation: Optional[str]):

    # Init
    time_passed = (datetime.datetime.now()-start_time).total_seconds()
    pct_progress = 100.0-(nb_todo-nb_done)*100/nb_todo
    
    # If we haven't really started yet, don't report time estimate yet
    if nb_done == 0:
        print(f"\r  ?: ? left to do {operation} on {(nb_todo-nb_done):8d} of {nb_todo:8d} ({pct_progress:3.2f}%)    ", 
              end="", flush=True)
    elif time_passed > 0:
        # Else, report progress properly...
        processed_per_hour = (nb_done/time_passed) * 3600
        hours_to_go = (int)((nb_todo - nb_done)/processed_per_hour)
        min_to_go = (int)((((nb_todo - nb_done)/processed_per_hour)%1)*60)
        pct_progress = 100.0-(nb_todo-nb_done)*100/nb_todo
        if pct_progress < 100:
            print(f"\r{hours_to_go:3d}:{min_to_go:2d} left to do {operation} on {(nb_todo-nb_done):8d} of {nb_todo:8d} ({pct_progress:3.2f}%)    ", 
                  end="", flush=True)
        else:
            print(f"\r{hours_to_go:3d}:{min_to_go:2d} left to do {operation} on {(nb_todo-nb_done):8d} of {nb_todo:8d} ({pct_progress:3.2f}%)    \n", 
                  end="", flush=True)


def formatbytes(bytes: float):
    """
    Return the given bytes as a human friendly KB, MB, GB, or TB string
    """

    bytes_float = float(bytes)
    KB = float(1024)
    MB = float(KB ** 2) # 1,048,576
    GB = float(KB ** 3) # 1,073,741,824
    TB = float(KB ** 4) # 1,099,511,627,776

    if bytes_float < KB:
        return '{0} {1}'.format(bytes_float,'Bytes' if 0 == bytes_float > 1 else 'Byte')
    elif KB <= bytes_float < MB:
        return '{0:.2f} KB'.format(bytes_float/KB)
    elif MB <= bytes_float < GB:
        return '{0:.2f} MB'.format(bytes_float/MB)
    elif GB <= bytes_float < TB:
        return '{0:.2f} GB'.format(bytes_float/GB)
    elif TB <= bytes_float:
        return '{0:.2f} TB'.format(bytes_float/TB)

ParallellisationParams = NamedTuple('result', [('nb_parallel', int), ('nb_batches_recommended', int), ('nb_rows_per_batch', int)])
def get_parallellisation_params(
        nb_rows_total: int,
        nb_parallel: int = -1,
        prev_nb_batches: int = None,
        verbose: bool = False) -> ParallellisationParams:
    """
    Determines recommended parallellisation params.

    Args:
        nb_rows_total (int): The total number of rows that will be processed
        nb_parallel (int, optional): The level of parallellisation requested. 
            If -1, tries to use all resources available. Defaults to -1.
        prev_nb_batches (int, optional): If applicable, the number of batches 
            used in a previous pass of the calculation. Defaults to None.
        verbose (bool, optional): [description]. Defaults to False.

    Returns:
        ParallellisationParams (NamedTuple('result', [('nb_parallel', int), ('nb_batches_recommended', int), ('nb_rows_per_batch', int)])): The recommended parameters.
    """
    # Some initialisations
    bytes_basefootprint = 50*1024*1024   # Base footprint of a python process
    bytes_per_row = 100                  # Average memory needed per row in bytes. Remark: when running from VS code, 3 times higher!
    min_avg_rows_per_batch = 1000
    max_avg_rows_per_batch = 7500       # 60.000 on small test, but seems slow in larger
    bytes_min_per_process = bytes_basefootprint + bytes_per_row * min_avg_rows_per_batch
    bytes_usable = psutil.virtual_memory().available * 0.9
    
    # If the number of rows is really low, just use one batch
    # TODO: for very complex features, possibly this limit is not a good idea
    if nb_rows_total < min_avg_rows_per_batch:
        return ParallellisationParams(1, 1, nb_rows_total)

    if(nb_parallel == -1):
        nb_parallel = multiprocessing.cpu_count()

    # If the available memory is very small, check if we can use more swap 
    if bytes_usable < 1024*1024:
        bytes_usable = min(psutil.swap_memory().free, 1024*1024)
    logger.debug(f"memory_usable: {formatbytes(bytes_usable)} with mem.available: {formatbytes(psutil.virtual_memory().available)} and swap.free: {formatbytes(psutil.swap_memory().free)}") 

    # If not enough memory for the amount of parallellism asked, reduce
    if (nb_parallel * bytes_min_per_process) > bytes_usable:
        nb_parallel = int(bytes_usable/bytes_min_per_process)
        logger.debug(f"Nb_parallel reduced to {nb_parallel} to evade excessive memory usage")

    # Optimal number of batches and rows per batch based on memory usage
    nb_batches = math.ceil((nb_rows_total*bytes_per_row*nb_parallel)/(bytes_usable-bytes_basefootprint*nb_parallel))
    
    # Make sure the average batch doesn't contain > max_avg_rows_per_batch
    batch_size = math.ceil(nb_rows_total/nb_batches)
    if batch_size > max_avg_rows_per_batch:
        batch_size = max_avg_rows_per_batch
        nb_batches = math.ceil(nb_rows_total/batch_size)
    mem_predicted = (bytes_basefootprint + batch_size*bytes_per_row)*nb_batches

    # Make sure there are enough batches to use as much parallelism as possible
    if nb_batches > 1 and nb_batches < nb_parallel:
        if prev_nb_batches is None:
            nb_batches = round(nb_parallel*1.25)
        elif nb_batches < prev_nb_batches/4:
            nb_batches = round(nb_parallel*1.25)
    
    batch_size = math.ceil(nb_rows_total/nb_batches)

    # Log result
    if verbose:
        logger.info(f"nb_batches_recommended: {nb_batches}, rows_per_batch: {batch_size} for nb_rows_input_layer: {nb_rows_total} will result in mem_predicted: {formatbytes(mem_predicted)}")   

    return ParallellisationParams(nb_parallel, nb_batches, batch_size)




import os
from pathlib import Path
import sys
sys.path.insert(0, str(Path(__file__).resolve().parent / '..'))

from geofileops.util import ogr_util
from geofileops import geofile

class GdalBin():
    def __init__(self, set_gdal_bin: bool = True, gdal_bin_path: str = None):
        self.set_gdal_bin = set_gdal_bin
        if set_gdal_bin is True:
            if gdal_bin_path is None:
                self.gdal_bin_path = r"X:\GIS\Software\_Progs\OSGeo4W64_2020-05-29\bin"
            else:
                self.gdal_bin_path = gdal_bin_path

    def __enter__(self):
        if self.set_gdal_bin is True:
            import os
            os.environ['GDAL_BIN'] = self.gdal_bin_path

    def __exit__(self, type, value, traceback):
        #Exception handling here
        import os
        if os.environ['GDAL_BIN'] is not None:
            del os.environ['GDAL_BIN']

def get_testdata_dir() -> Path:
    return Path(__file__).resolve().parent / 'data'

def is_gdal_default_ok() -> bool:
    return False

def is_gdal_bin_ok() -> bool:
    return True

def test_get_gdal_to_use():

    # On windows, the default gdal installation with conda doesn't work
    if os.name == 'nt': 
        # If GDAL_BIN not set, should be not OK
        try:
            ogr_util.get_gdal_to_use('ST_area()')
            test_ok = True
        except:
            test_ok = False
        assert test_ok is is_gdal_default_ok(), "On windows, check is expected to be OK if GDAL_BIN is not set"

        # If GDAL_BIN set, it should be ok as well
        with GdalBin():
            try:
                ogr_util.get_gdal_to_use('ST_area()')
                test_ok = True
            except:
                test_ok = False
            assert test_ok is is_gdal_bin_ok(), "On windows, check is expected to be OK if GDAL_BIN is set (properly)"
        
    else:
        try:
            ogr_util.get_gdal_to_use('ST_area()')
            test_ok = True
        except:
            test_ok = False
        assert test_ok is True, "If not on windows, check is expected to be OK without setting GDAL_BIN"

if __name__ == '__main__':
    import tempfile
    tmpdir = tempfile.gettempdir()
    test_get_gdal_to_use()



import os
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

version = "0.0.10"
with open('package_tag_version.txt', mode='w') as file:
    file.write(version)

setuptools.setup(
    name="geofileops", 
    version=version,
    author="Pieter Roggemans",
    author_email="pieter.roggemans@gmail.com",
    description="Package to do spatial operations on geo files.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/theroggy/geofileops",
    packages=setuptools.find_packages(),
    install_requires=["geopandas>=0.8", "pygeos", "psutil"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)


# -*- coding: utf-8 -*-
"""
Module with helper functions for geo files.
"""

import datetime
import filecmp
import logging
import os
from pathlib import Path
import pyproj
import shutil
import time
from typing import Any, List, Optional, Tuple, Union

import fiona
import geopandas as gpd
from osgeo import gdal

from .util import io_util
from .util import ogr_util

#-------------------------------------------------------------
# First define/init some general variables/constants
#-------------------------------------------------------------
# Get a logger...
logger = logging.getLogger(__name__)
#logger.setLevel(logging.DEBUG)

shapefile_suffixes = ['.shp', '.dbf', '.shx', '.prj', '.qix', '.sbn', '.sbx']
gdal.UseExceptions()        # Enable exceptions

#-------------------------------------------------------------
# The real work
#-------------------------------------------------------------

def listlayers(
        path: Union[str, 'os.PathLike[Any]'],
        verbose: bool = False) -> List[str]:

    """
    Get the layers in a geofile.

    Args:
        path (PathLike): path to the file to get info about
        verbose (bool, optional): True to enable verbose logging. Defaults to False.

    Returns:
        List[str]: the list of layers
    """
    return fiona.listlayers(str(path))

class LayerInfo:
    def __init__(self, 
            name: str,
            featurecount: int, 
            total_bounds: Tuple[float, float, float, float],
            geometrycolumn: str, 
            geometrytypename: str,
            columns: List[str],
            crs: pyproj.CRS):
        self.name = name
        self.featurecount = featurecount
        self.total_bounds = total_bounds
        self.geometrycolumn = geometrycolumn
        self.geometrytypename = geometrytypename
        self.columns = columns
        self.crs = crs

    def __repr__(self):
        return f"{self.__class__}({self.__dict__})"
    
def getlayerinfo(
        path: Union[str, 'os.PathLike[Any]'],
        layer: str = None,
        verbose: bool = False) -> LayerInfo:
    """
    Get information about a layer in the geofile.

    Args:
        path (PathLike): path to the file to get info about
        layer (str): the layer you want info about. Doesn't need to be 
            specified if there is only one layer in the geofile. 
        verbose (bool, optional): True to enable verbose logging. Defaults to False.

    Returns:
        LayerInfo: the information about the layer
    """        
    ##### Init #####
    datasource = gdal.OpenEx(str(path))
    if layer is not None:
        datasource_layer = datasource.GetLayer(layer)
    elif datasource.GetLayerCount() == 1:
        datasource_layer = datasource.GetLayerByIndex(0)
    else:
        raise Exception(f"No layer specified, and file has <> 1 layer: {path}")

    # If the layer doesn't exist, return 
    if datasource_layer is None:
        raise Exception(f"Layer {layer} not found in file: {path}")

    # Get column info
    columns = []
    layer_defn = datasource_layer.GetLayerDefn()
    for i in range(layer_defn.GetFieldCount()):
        columns.append(layer_defn.GetFieldDefn(i).GetName())
    geometrycolumn = datasource_layer.GetGeometryColumn()
    if geometrycolumn == '':
        geometrycolumn = 'geometry'
    geometrytypename = gdal.ogr.GeometryTypeToName(datasource_layer.GetGeomType())
    geometrytypename = geometrytypename.replace(' ', '').upper()
    
    # For shape files, the geometrytypename is sometimes wrong...
    if Path(path).suffix.lower() == '.shp' and geometrytypename == 'LINESTRING':
        geometrytypename = 'MULTILINESTRING'       
    
    # Convert gdal extent (xmin, xmax, ymin, ymax) to bounds (xmin, ymin, xmax, ymax)
    extent = datasource_layer.GetExtent()
    total_bounds = (extent[0], extent[2], extent[1], extent[3])

    # Get projection
    crs = None
    spatialref = datasource_layer.GetSpatialRef()
    if spatialref is not None:
        crs = pyproj.CRS(spatialref.ExportToWkt())

    return LayerInfo(
            name=datasource_layer.GetName(),
            featurecount=datasource_layer.GetFeatureCount(),
            total_bounds=total_bounds,
            geometrycolumn=geometrycolumn, 
            geometrytypename=geometrytypename,
            columns=columns,
            crs=crs)

def get_only_layer(path: Union[str, 'os.PathLike[Any]']) -> str:
    """
    Get the layername for a file that only contains one layer.

    If the file contains multiple layers, an exception is thrown.

    Args:
        path (PathLike): the file

    Raises:
        Exception: [description]
        Exception: [description]

    Returns:
        str: the layer name
    """
    layers = fiona.listlayers(str(path))
    nb_layers = len(layers)
    if nb_layers == 1:
        return layers[0]
    elif nb_layers == 0:
        raise Exception(f"Error: No layers found in {path}")
    else:
        raise Exception(f"Error: More than 1 layer found in {path}: {layers}")

def get_default_layer(path: Union[str, 'os.PathLike[Any]']):
    return Path(path).stem

def has_spatial_index(
        path: Union[str, 'os.PathLike[Any]'],
        layer: str = None,
        geometrycolumn: str = None) -> bool:
    # Init
    path_p = Path(path)

    # Now check the index
    driver = get_driver(path_p)    
    if driver == 'GPKG':
        layerinfo = getlayerinfo(path_p, layer)
        data_source = gdal.OpenEx(str(path_p), nOpenFlags=gdal.OF_READONLY)
        result = data_source.ExecuteSQL(
                f"SELECT HasSpatialIndex('{layerinfo.name}', '{layerinfo.geometrycolumn}')",
                dialect='SQLITE')
        row = result.GetNextFeature()
        has_spatial_index = row.GetField(0)
        return (has_spatial_index == 1)
    elif driver == 'ESRI Shapefile':
        index_path = path_p.parent / f"{path_p.stem}.qix" 
        return index_path.exists()
    else:
        raise Exception(f"has_spatial_index not supported for {path_p}")

def create_spatial_index(
        path: Union[str, 'os.PathLike[Any]'],
        layer: str = None):
    # Init
    path_p = Path(path)
    layerinfo = getlayerinfo(path_p, layer)

    # Now really add index
    datasource = gdal.OpenEx(str(path_p), nOpenFlags=gdal.OF_UPDATE)
    driver = get_driver(path_p)    
    if driver == 'GPKG':
        datasource.ExecuteSQL(
                f"SELECT CreateSpatialIndex('{layerinfo.name}', '{layerinfo.geometrycolumn}')",                
                dialect='SQLITE') 
    else:
        datasource.ExecuteSQL(f'CREATE SPATIAL INDEX ON "{layerinfo.name}"')

def remove_spatial_index(
        path: Union[str, 'os.PathLike[Any]'],
        layer: str = None):
    # Init
    path_p = Path(path)
    layerinfo = getlayerinfo(path_p, layer)

    # Now really remove index
    datasource = gdal.OpenEx(str(path_p), nOpenFlags=gdal.OF_UPDATE)
    driver = get_driver(path_p)    
    if driver == 'GPKG':
        datasource.ExecuteSQL(
                f"SELECT DisableSpatialIndex('{layerinfo.name}', '{layerinfo.geometrycolumn}')",
                dialect='SQLITE') 
    elif driver == 'ESRI Shapefile':
        # DROP SPATIAL INDEX ON ... command gives an error, so just remove .qix
        index_path = path_p.parent / f"{path_p.stem}.qix" 
        index_path.unlink()
    else:
        datasource.ExecuteSQL(f'DROP SPATIAL INDEX ON "{layerinfo.name}"')

def rename_layer(
        path: Union[str, 'os.PathLike[Any]'],
        layer: Optional[str],
        new_layer: str):
    # Check input parameters
    path_p = Path(path)
    if layer is None:
        layer = get_only_layer(path_p)

    # Now really rename
    datasource = gdal.OpenEx(str(path_p), nOpenFlags=gdal.OF_UPDATE)
    sql_stmt = f'ALTER TABLE "{layer}" RENAME TO "{new_layer}"'
    datasource.ExecuteSQL(sql_stmt)

def add_column(
        path: Union[str, 'os.PathLike[Any]'],
        name: str,
        type: str,
        expression: str = None, 
        layer: str = None,
        force_update: bool = False):
    """
    Add a column to a layer of the geofile.

    Args:
        path (PathLike): Path to the geofile
        name (str): Name for the new column
        type (str): Column type.
        expression (str, optional): SQLite expression to use to update 
            the value. Defaults to None.
        layer (str, optional): The layer name. If None and the geofile
            has only one layer, that layer is used. Defaults to None.
        force_update (bool, optional): If the column already exists, execute 
            the update anyway. Defaults to False. 

    Raises:
        ex: [description]
    """

    ##### Init #####
    if type.upper() not in ['TEXT', 'NUMERIC', 'INTEGER', 'REAL', 'BLOB']:
        raise Exception(f"Type specified is not supported: {type}")
    path_p = Path(path)
    name = name.lower()
    if layer is None:
        layer = get_only_layer(path_p)
    layerinfo_orig = getlayerinfo(path_p, layer)
    
    ##### Go! #####
    datasource = None
    try:
        #datasource = gdal.OpenEx(str(path_p), nOpenFlags=gdal.OF_UPDATE)
        if name not in layerinfo_orig.columns:
            # If column doesn't exist yet, create it
            #if name not in getlayerinfo(path_p, layer=layer).columns:
            sqlite_stmt = f'ALTER TABLE "{layer}" ADD COLUMN "{name}" {type}'
            ogr_util.vector_info(path=path_p, sql_stmt=sqlite_stmt, sql_dialect='SQLITE', readonly=False)
            #datasource.ExecuteSQL(sqlite_stmt, dialect='SQLITE')
        else:
            logger.warning(f"Column {name} existed already in {path_p}, layer {layer}")
            
        # If an expression was provided and update can be done, go for it...
        if(expression is not None 
           and (name not in layerinfo_orig.columns 
                or force_update is True)):
            sqlite_stmt = f'UPDATE "{layer}" SET "{name}" = {expression}'
            ogr_util.vector_info(path=path_p, sql_stmt=sqlite_stmt, sql_dialect='SQLITE', readonly=False)
            #datasource.ExecuteSQL(sqlite_stmt, dialect='SQLITE')
    finally:
        if datasource is not None:
            del datasource

def read_file(
        path: Union[str, 'os.PathLike[Any]'],
        layer: str = None,
        columns: List[str] = None,
        bbox = None,
        rows = None,
        ignore_geometry: bool = False) -> gpd.GeoDataFrame:
    """
    Reads a file to a pandas dataframe. The fileformat is detected based on the filepath extension.

    # TODO: think about if possible/how to support adding optional parameter and pass them to next function, example encoding, float_format,...
    """
    # Init
    path_p = Path(path)
    ext_lower = path_p.suffix.lower()

    # For file multilayer types, if no layer name specified, check if there is only one layer in the file.
    if(ext_lower in ['.gpkg'] 
       and layer is None):
        listlayers = fiona.listlayers(str(path_p))
        if len(listlayers) == 1:
            layer = listlayers[0]
        else:
            raise Exception(f"File contains {len(listlayers)} layers: {listlayers}, but layer is not specified: {path}")

    # Depending on the extension... different implementations
    if ext_lower == '.shp':
        result_gdf = gpd.read_file(str(path_p), bbox=bbox, rows=rows, ignore_geometry=ignore_geometry)
    elif ext_lower == '.geojson':
        result_gdf = gpd.read_file(str(path_p), bbox=bbox, rows=rows, ignore_geometry=ignore_geometry)
    elif ext_lower == '.gpkg':
        result_gdf = gpd.read_file(str(path_p), layer=layer, bbox=bbox, rows=rows, ignore_geometry=ignore_geometry)
    else:
        raise Exception(f"Not implemented for extension {ext_lower}")

    # If columns to read are specified... filter 
    if columns is not None:
        result_gdf = result_gdf[columns]
        
    return result_gdf

def to_file(
        gdf: gpd.GeoDataFrame,
        path: Union[str, 'os.PathLike[Any]'],
        layer: str = None,
        append: bool = False,
        append_timeout_s: int = 600,
        index: bool = True):
    """
    Writes a pandas dataframe to file. The fileformat is detected based on the filepath extension.
    """
    # TODO: think about if possible/how to support adding optional parameter and pass them to next 
    # function, example encoding, float_format,...

    # Check input parameters
    path_p = Path(path)

    # If no layer name specified, use the filename (without extension)
    if layer is None:
        layer = Path(path_p).stem
    # If the dataframe is empty, log warning and return
    if len(gdf) <= 0:
        #logger.warn(f"Cannot write an empty dataframe to {filepath}.{layer}")
        return

    def write_to_file(
            gdf: gpd.GeoDataFrame,
            path: Path, 
            layer: str, 
            index: bool = True,
            append: bool = False):

        if append is True:
            if path.exists():
                mode = 'a'
            else:
                mode = 'w'
        else:
            mode = 'w'

        ext_lower = path.suffix.lower()
        if ext_lower == '.shp':
            if index is True:
                gdf = gdf.reset_index(inplace=False)
            gdf.to_file(str(path), mode=mode)
        elif ext_lower == '.gpkg':
            gdf.to_file(str(path), layer=layer, driver="GPKG", mode=mode)
        else:
            raise Exception(f"Not implemented for extension {ext_lower}")

    # If no append, just write to output path
    if append is False:
        write_to_file(gdf=gdf, path=path_p, layer=layer, index=index, append=False)
    else:
        # If append is asked, check if the fiona driver supports appending. If
        # not, write to temporary output file 
        driver = get_driver(path_p)
        gdftemp_path = None
        gdftemp_lockpath = None
        if 'a' not in fiona.supported_drivers[driver]:
            # Get a unique temp file path. The file cannot be created yet, so 
            # only create a lock file to evade other processes using the same 
            # temp file name 
            gdftemp_path, gdftemp_lockpath = io_util.get_tempfile_locked(
                    base_filename='gdftemp',
                    suffix=path_p.suffix,
                    dirname='geofile_to_file')
            write_to_file(gdf, path=gdftemp_path, layer=layer, index=index)  
        
        # Files don't typically support having multiple processes writing 
        # simultanously to them, so use lock file to synchronize access.
        lockfile = Path(f"{str(path_p)}.lock")
        start_time = datetime.datetime.now()
        while(True):
            if io_util.create_file_atomic(lockfile) is True:
                try:
                    # If gdf wasn't written to temp file, use standard write-to-file
                    if gdftemp_path is None:
                        write_to_file(gdf=gdf, path=path_p, layer=layer, index=index, append=True)
                    else:
                        # If gdf was written to temp file, use append_to_nolock + cleanup
                        _append_to_nolock(src=gdftemp_path, dst=path_p, dst_layer=layer)
                        remove(gdftemp_path)
                        gdftemp_lockpath.unlink()
                finally:
                    lockfile.unlink()
                    return
            else:
                time_waiting = (datetime.datetime.now()-start_time).total_seconds()
                if time_waiting > append_timeout_s:
                    raise Exception(f"append_to_layer timeout of {append_timeout_s} reached, so stop trying!")
            
            # Sleep for a second before trying again
            time.sleep(1)
        
def get_crs(path: Union[str, 'os.PathLike[Any]']) -> pyproj.CRS:
    with fiona.open(str(path), 'r') as geofile:
        return pyproj.CRS(geofile.crs)

def is_geofile(path: Union[str, 'os.PathLike[Any]']) -> bool:
    """
    Determines based on the filepath if this is a geofile.
    """
    return is_geofile_ext(Path(path).suffix)

def is_geofile_ext(file_ext: str) -> bool:
    """
    Determines based on the file extension if this is a geofile.
    """
    file_ext_lower = file_ext.lower()
    if file_ext_lower in ('.shp', '.gpkg', '.geojson'):
        return True
    else:
        return False

def cmp(
        path1: Union[str, 'os.PathLike[Any]'], 
        path2: Union[str, 'os.PathLike[Any]']) -> bool:
    """
    Compare if two geofiles are identical. 

    For geofiles that use multiple files, all relevant files must be identical.
    Eg. for shapefiles, the .shp, .shx and .dbf file must be identical.

    Args:
        path1 (PathLike): path to the first file
        path2 (PathLike): path to the second file

    Returns:
        bool: True if the files are identical
    """
    # Check input parameters
    path1_p = Path(path1)
    path2_p = Path(path2)

    # For a shapefile, multiple files need to be compared
    if path1_p.suffix.lower() == '.shp':
        path2_noext, _ = os.path.splitext(path2_p)
        shapefile_base_suffixes = [".shp", ".dbf", ".shx"]
        path1_noext = path1_p.parent / path1_p.stem
        path2_noext = path2_p.parent / path2_p.stem
        for ext in shapefile_base_suffixes:
            if not filecmp.cmp(f"{str(path1_noext)}{ext}", f"{str(path2_noext)}{ext}"):
                logger.info(f"File {path1_noext}{ext} is different from {path2_noext}{ext}")
                return False
        return True
    else:
        return filecmp.cmp(str(path1_p), str(path2_p))
    
def copy(
        src: Union[str, 'os.PathLike[Any]'], 
        dst: Union[str, 'os.PathLike[Any]']):
    """
    Copies the geofile from src to dst. Is the source file is a geofile containing
    of multiple files (eg. .shp) all files are copied.

    Args:
        src (PathLike): the file to copy
        dst (PathLike): the location to copy the file(s) to
    """
    # Check input parameters
    src_p = Path(src)
    dst_p = Path(dst)

    # For a shapefile, multiple files need to be copied
    if src_p.suffix.lower() == '.shp':
        # If dest is a dir, just use move. Otherwise concat dest filepaths
        if dst_p.is_dir():
            for ext in shapefile_suffixes:
                srcfile = src_p.parent / f"{src_p.stem}{ext}"
                if srcfile.exists():
                    shutil.copy(str(srcfile), dst_p)
        else:
            for ext in shapefile_suffixes:
                srcfile = src_p.parent / f"{src_p.stem}{ext}"
                dstfile = dst_p.parent / f"{dst_p.stem}{ext}"
                if srcfile.exists():
                    shutil.copy(str(srcfile), dstfile)                
    else:
        return shutil.copy(str(src_p), dst_p)

def move(
        src: Union[str, 'os.PathLike[Any]'], 
        dst: Union[str, 'os.PathLike[Any]']):
    """
    Moves the geofile from src to dst. Is the source file is a geofile containing
    of multiple files (eg. .shp) all files are moved.

    Args:
        src (PathLike): the file to move
        dst (PathLike): the location to move the file(s) to
    """
    # Check input parameters
    src_p = Path(src)
    dst_p = Path(dst)

    # For a shapefile, multiple files need to be copied
    if src_p.suffix.lower() == '.shp':
        # If dest is a dir, just use move. Otherwise concat dest filepaths
        if dst_p.is_dir():
            for ext in shapefile_suffixes:
                srcfile = src_p.parent / f"{src_p.stem}{ext}"
                if srcfile.exists():
                    shutil.move(str(srcfile), dst_p, copy_function=io_util.copyfile)
        else:
            for ext in shapefile_suffixes:
                srcfile = src_p.parent / f"{src_p.stem}{ext}"
                dstfile = dst_p.parent / f"{dst_p.stem}{ext}"
                if srcfile.exists():
                    shutil.move(str(srcfile), dstfile, copy_function=io_util.copyfile)                
    else:
        return shutil.move(str(src_p), dst_p, copy_function=io_util.copyfile)

def remove(path: Union[str, 'os.PathLike[Any]']):
    """
    Removes the geofile. Is it is a geofile composed of multiple files 
    (eg. .shp) all files are removed.

    Args:
        path (PathLike): the file to remove
    """
    # Check input parameters
    path_p = Path(path)

    # For a shapefile, multiple files need to be copied
    if path_p.suffix.lower() == '.shp':
        for ext in shapefile_suffixes:
            curr_path = path_p.parent / f"{path_p.stem}{ext}"
            if curr_path.exists():
                curr_path.unlink()
    else:
        path_p.unlink()

def append_to(
        src: Union[str, 'os.PathLike[Any]'], 
        dst: Union[str, 'os.PathLike[Any]'],
        src_layer: str = None,
        dst_layer: str = None,
        force_output_geometrytype: str = None,
        create_spatial_index: bool = True,
        append_timeout_s: int = 600,
        verbose: bool = False):
    """
    Append is not supported for all filetypes in fiona/geopandas (0.8)... so 
    workaround via gdal needed.
    """
    src_p = Path(src)
    dst_p = Path(dst)
    
    # Files don't typically support having multiple processes writing 
    # simultanously to them, so use lock file to synchronize access.
    lockfile = Path(f"{str(dst_p)}.lock")
    start_time = datetime.datetime.now()
    while(True):
        if io_util.create_file_atomic(lockfile) is True:
            try:
                # append
                _append_to_nolock(
                        src=src_p, 
                        dst=dst_p,
                        src_layer=src_layer,
                        dst_layer=dst_layer,
                        force_output_geometrytype=force_output_geometrytype,
                        create_spatial_index=create_spatial_index,
                        verbose=verbose)
            finally:
                lockfile.unlink()
                return
        else:
            time_waiting = (datetime.datetime.now()-start_time).total_seconds()
            if time_waiting > append_timeout_s:
                raise Exception(f"append_to_layer timeout of {append_timeout_s} reached, so stop trying!")
        
        # Sleep for a second before trying again
        time.sleep(1)

def _append_to_nolock(
        src: Path, 
        dst: Path,
        src_layer: str = None,
        dst_layer: str = None,
        force_output_geometrytype: str = None,
        create_spatial_index: bool = True,
        verbose: bool = False):
    # append
    translate_info = ogr_util.VectorTranslateInfo(
            input_path=src,
            output_path=dst,
            translate_description=None,
            input_layers=src_layer,
            output_layer=dst_layer,
            transaction_size=200000,
            append=True,
            update=True,
            force_output_geometrytype=force_output_geometrytype,
            create_spatial_index=create_spatial_index,
            priority_class='NORMAL',
            force_py=True,
            verbose=verbose)
    ogr_util.vector_translate_by_info(info=translate_info)

def get_driver(path: Union[str, 'os.PathLike[Any]']) -> str:
    """
    Get the driver to use for the file extension of this filepath.
    """
    return get_driver_for_ext(Path(path).suffix)

def get_driver_for_ext(file_ext: str) -> str:
    """
    Get the driver to use for this file extension.
    """
    file_ext_lower = file_ext.lower()
    if file_ext_lower == '.shp':
        return 'ESRI Shapefile'
    elif file_ext_lower == '.geojson':
        return 'GeoJSON'
    elif file_ext_lower == '.gpkg':
        return 'GPKG'
    else:
        raise Exception(f"Not implemented for extension {file_ext_lower}")        

def to_multi_type(geometrytypename: str) -> str:
    """
    Map the input geometry type to the corresponding 'MULTI' geometry type...

    Args:
        geometrytypename (str): Input geometry type

    Raises:
        Exception: If input geometrytype is not known.

    Returns:
        str: Corresponding 'MULTI' geometry type
    """
    if geometrytypename in ['MULTIPOINT', 'MULTILINESTRING', 'MULTIPOLYGON', 'GEOMETRYCOLLECTION']:
        return geometrytypename
    elif geometrytypename in ['POINT', 'LINESTRING', 'POLYGON']:
        return f"MULTI{geometrytypename}"
    elif geometrytypename == 'GEOMETRY':
        return geometrytypename
    else:
        raise Exception(f"Unknown geometrytype: {geometrytypename}")

def to_generaltypeid(geometrytypename: str) -> int:
    """
    Map the input geometry type to the corresponding geometry type id:
        * 1 = POINT-type
        * 2 = LINESTRING-type
        * 3 = POLYGON-type

    Args:
        geometrytypename (str): Input geometry type

    Raises:
        Exception: If input geometrytype is not known.

    Returns:
        int: Corresponding geometry type id
    """
    if geometrytypename in ['POINT', 'MULTIPOINT']:
        return 1
    elif geometrytypename in ['LINESTRING', 'MULTILINESTRING']:
        return 2
    elif geometrytypename in ['POLYGON', 'MULTIPOLYGON']:
        return 3
    else:
        raise Exception(f"Unsupported geometrytype ({geometrytypename})")




import json
import logging
import logging.config
import os
from pathlib import Path
import sys
sys.path.insert(0, str(Path(__file__).resolve().parent / '..'))

import geofileops.geofileops as geofileops

if __name__ == '__main__':

    ##### Init #####
    # Init logging
    script_dir = Path(__file__).resolve().parent
    with open(script_dir / 'logging.json', 'r') as log_config_file:
        log_config_dict = json.load(log_config_file)
    logging.config.dictConfig(log_config_dict)
    logger = logging.getLogger()

    # Use OSGeo4W for ogr operations
    if 'GDAL_BIN' not in os.environ:
        os.environ['GDAL_BIN'] = r"X:\GIS\Software\_Progs\OSGeo4W64_2020-05-29\bin"
        os.environ['PATH'] = r"X:\GIS\Software\_Progs\OSGeo4W64_2020-05-29\bin;" + os.environ['PATH']

    input1_path = Path(r"X:\PerPersoon\PIEROG\Taken\2020\2020-07-23_rbh2010vsrbh2019\rbh2019_ms.gpkg")
    input2_path = Path(r"X:\PerPersoon\PIEROG\Taken\2020\2020-07-23_rbh2010vsrbh2019\rbh2010_ms.gpkg")

    output_filename = f"{input1_path.stem}_INTER_{input2_path.stem}.gpkg"
    output_path = input1_path.parent / output_filename

    ##### Go! #####
    logger.info("Start")
    geofileops.intersect(
            input1_path=input1_path,
            input2_path=input2_path,
            output_path=output_path,
            force=True)
    logger.info("Ready")
